/*
 Code for picoverter board.
 two ntc 10K thermistors, one on the heatsink, one on the toroid
 run/stop switch
 Over temp inverter stop function 
 Under voltage stop function
 fan control with hyteresis
 20x4 LCD, serial data out A4 pin at 10,000 baud

*/
#define HS_FAN_ON      45       // deg.C
#define HS_FAN_OFF     35       // all 6 temperatures need to be less than 100 deg.C
#define HS_STOP        55       // due to lookup table based voltage to temperature code

#define TOR_FAN_ON     45       // have fan on temp rather more than fan off
#define TOR_FAN_OFF    35       // to ensure the fan runs for a decent time
#define TOR_STOP       55       // and gets the job done.

#define NTC_COMBINE     0       // = 0 if use two fans, = 1 to use one fan only, the HS fan

#define LV_CUT_OFF    46.0      // minimum battery volts for inverter to run
#define LV_RESTART    51.0      // restart volts
#define LV_STOP_ENABLE  0       //  = 1 to allow LV cut-off, = 0 to just keep running no matter battery voltage
#define LV_ENABLE     1         // = 1 to permit LV restart if battery increases enough, = 0 to disable
#define LV_TIMEOUT    5        // how many seconds BV needs to be below LV_CUT_OFF. Noise filter  

#define ADC_OUT  0          // channel for AC output sample
#define HS_NTC   1          // and others
#define TOR_NTC  2
#define BV       3

#define BV_SCALE 0.05986       // or something. Set it via trial and error.

#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))       // way, way faster than digitalWrite()
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

#define NPWM  200         // 200 for this and 800 for PPWM for 50Hz and 20kHz PWM. 
#define PPWM  800         // 50.002 Hz 799  49.939 Hz for 800

uint8_t hs_temp,tor_temp;
float bv;
int lv_time_count, lv_sec_count;
    
uint8_t uf,oen,v1low,lv_stop,ot_stop,led_count;
uint16_t pcount;
uint32_t vpwr;
int sst;
uint16_t l[NPWM+1];
float pwr;
uint8_t dbounce;
double Input, Output, Setpoint;
double errSum, lastErr,error,dErr;
double kp, ki, kd;
float ac_output, ac_setpoint;

void init_vars(void);
float get_temp(int);
uint8_t lookup_temp(int);

uint8_t init_done; 
volatile uint8_t ntc[675];

volatile char buf[85],*bp,bcount;
volatile uint8_t blcd,go_lcd = 0,bpos=0;
volatile int baudc = 0,baud_waits;

void setup()
  {
  int i;
  float t,u;
  for(t=0.0,i=0; i <= NPWM; i++,t +=   3.14159/ (float) (NPWM+1))
    {   
    u = 65535.0 * sin(t);   // 1/2 wave sine lookup table, scaled to (16 bits - 1)
    l[i] =  (int)u;  
    }

  for(i=270; i < 944; i++)
    {
      t = get_temp(i);
      ntc[i-270] = (int)t;
    }
  for(i = 0; i < 80; i++)
    buf[i] = ' ';
      
  pinMode(9, OUTPUT);     // V1 drive
  pinMode(10, OUTPUT);    // V2 drive
  pinMode(12, OUTPUT);    // inverter ON/OFF LED, pin 15 on nano, Arduino D12
  pinMode(13, OUTPUT);    // over temp stop and lv stop indicator
  pinMode(8,INPUT);       // continuous on/off pin, driven by nano2
  pinMode(5,OUTPUT);      // used for IR2110 shutdown drive, LOW = shutdown
  pinMode(7,OUTPUT);      // DSO sync pulse, each 50Hz
  pinMode(3,OUTPUT);      // heat sink fan MOSFET gate drive
  pinMode(4,OUTPUT);      // toroid fan drive
  pinMode(18,OUTPUT);     // my bit-banged serial port, 10000 baud, uses pin A4
  noInterrupts();

  TCCR1A = _BV(COM1A1) | _BV(COM1B0) | _BV(COM1B1) | _BV(WGM11);
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);
  OCR1A = 10;
  ICR1 = PPWM;
  TIMSK1 |= (1 << TOIE1);

  cbi(PORTD,5);   // disable gate drive output. data direction flag set by above statement.
  oen = 0;        // output enable. = 1 to run inverter, = 0 to stop inverter
  init_vars();

  interrupts();             // enable all interrupts
  TIMSK0 = 0;               // except arduino clock

  sbi(ADCSRA,ADPS2) ;   // ADC clock prescale = 16 now. this saves 100us in main loop with little cost to loop stability
  cbi(ADCSRA,ADPS1) ;
  cbi(ADCSRA,ADPS0) ;  

  //Serial.begin(115200);     // remove in production code
  }
 
//
// 20Khz SPWM, code has only 50uS to run, takes about 10uS
//
ISR(TIMER1_OVF_vect)        
 {   
  long c;
  c = (l[pcount] * vpwr) >> 16;   // scale sine wave by vpwr, 32 bit integer calcs. 16 bit shift fastest of all
  if (v1low == 1)                 // alternate between 2 output compare pins. Get a full 50Hz waveform
    OCR1B = c;                    //  1/2 out this pin
  else
    OCR1A = c;                    // and then 1/2 out this pin. 
  pcount ++;
  if(pcount >= NPWM)          // pcount  will = NPWM, at each 1/2 wave, at the start
    {
    pcount = 0;               // reset counter
    uf=1;                     // enable one PID control loop execution, in loop()
    if (v1low == 1)           // if first 1/2 wave..
      {
       v1low = 0;             // toggle to 2nd half wave
       TCCR1A =  _BV(COM1A1) | _BV(WGM11); // config output compare to suit
       sbi(PORTD,7);          //DSO pulse 
      }                        
    else
      {
       v1low = 1;             // toggle to 1st half wave
       TCCR1A =  _BV(COM1B1) | _BV(WGM11);   // config O.C. to suit
       cbi(PORTD,7); 
      }
    }

  // send one char out pin A4. First, get a character and put it in "blcd" and init state machine.
if (go_lcd == 0 && baudc > 50)    // not send as soon as possible, let the LCD digest the data with a delay here
  {
  // get next char
  blcd = buf[bcount];
  bcount++;
  if (bcount > 80)
    {
    bcount = 0;
    blcd =  '$';     // this is translated by the serial LCD as "put cursor at top left position"
    }   
  go_lcd = 1;
  bpos = 0;
  baudc = 2;
  }
  
if (go_lcd == 1 && baudc > baud_waits)   // a little 8 bit serial output state machine
  {
    if(bpos == 9)
    {
    // send stop bit, pull high
    sbi(PORTC,4);
    go_lcd = 0;
    baudc = 0;
    bpos = 0;
    }
  if (bpos > 0  )
    {
      if (blcd & 1)
        sbi(PORTC,4);
      else
        cbi(PORTC,4);
      blcd = blcd >> 1;
      bpos++;  
      baudc = 0;
    }
  if(bpos == 0 && go_lcd == 1)
    {
    // send start bit, pull low
    cbi(PORTC,4);
    bpos++;  
    baudc = 0;
    }
  }
  baudc++;              // a counter, goes from 0, 1, 0 , 1 etc. to halve the output frequency of the serial data    
}


void init_vars()      // called prior to setting oen = 1, as well as from setup()
  {
  go_lcd = 0;     // the bit-bang serial port setup
  baud_waits = 1;
  baudc = 0;
  bcount = 0;             
  bp = &buf[0]; 
  
  ICR1 = PPWM;
  lv_stop = 0;  
  ot_stop = 0;
  sst = 0;      // slow start counter 0 - 251
  v1low = 1;    // half wave output flag, 0 or 1
  pcount = 0;   // counter for sine wave lookup table
  uf = 0;       // set to 1 to enable AC output voltage feedback code to execute
  pwr = 0.0;    // PWM duty cycle analog, 0.0 to 1.0
  vpwr = 0;      // integer equivalent of pwr, 0 to PPWM
  dbounce = 0;  // debounce counter for input switch 
  
  kp = 0.1;     // set up PID for AC output voltage control
  ki = 0.0;
  kd = 0.01;
  errSum = 0.0;        
  lastErr = 0.0;
  init_done = 1; 
  lv_sec_count = 0;
  lv_time_count = 0;
  led_count = 0;
  }


void do_pid()
{   
   double timeChange = 0.01;    // always called at 100Hz, so dT will be 10msec..
   error = Setpoint - Input;
   errSum += (error * timeChange);
   dErr = (error - lastErr) / timeChange;
   Output = kp * error + ki * errSum + kd * dErr;
   lastErr = error;
}

// loop() execution times:
// when uf = 1, i.e. zero cross of the AC waveform, it needs about 600uS to run
// when uf = 0, it needs 100 to 130 us to run
//
void loop()
{
  // sample channels
  ac_output = (float)analogRead(ADC_OUT)* 0.9765E-3; // scale 0 - 5V to 0.0 to 1.0
  hs_temp = (hs_temp * 7 + lookup_temp(analogRead(HS_NTC))) / 8;
  tor_temp = (tor_temp * 7 + lookup_temp(analogRead(TOR_NTC)) ) / 8;
  bv = bv * 0.999 + 0.001 * (float)analogRead(BV) * BV_SCALE;       // scale to volts
  
  if(uf == 1  )    // this will execute at 100Hz
      {            // it runs at near zero volt crossing of AC output       
      uf=0; 
      ac_setpoint = 0.55;       // for same Vfb as EG8010. change AC output voltage via Vfb voltage divider trimpot
      if (oen == 1) sst++;      // slow start. If output is enabled, keep starting up. sst = output pwm% or approx. voltage.
      if (sst > 251)sst = 251;  // stop increasing when sst has got to the end of slow start. when = 251, this enables PID. See below..
      if (oen == 0) sst--;      // if stopping, slow stop
      if (sst <= 0)             // once fully stopped, no more changes to sst needed
        {
        sst = 0;
        cbi(PORTD,5);     // pull IR2184 shutdown LOW to disable gate drive output
        }
      else
        sbi(PORTD,5);     // pull it HIGH, to enable output

      if(sst > 0 && sst < 250)     // slow start timer counts to 250 with gentle ramp up, then switches over to PID control. (gulp!)
        {                          // sst can increase or decrease, giving the slow start and slow stop.
        if (oen == 1)              // slow stop? I think of it as de-gauss
          { 
          if(ac_output < ac_setpoint)
            pwr += 1.0/250.0;     // bang-bang control, should AC output reach setpoint prior to PID.
          else                    // During slow start need to ensure output is tracking setpoint
            pwr -= 1.0/250.0;     // should output get close enough to matter.
          }
        else
          pwr -= 1.0 /250.0;      // this will execute in slow stop
        }
        
      if (sst == 251)       // once sst = 251, run PID. the above bang-bang control will NOT have executed, because sst = 251
        {                   
        Setpoint = ac_setpoint;     
        Input = ac_output;  
        do_pid();
        pwr = pwr + Output;
        }
        
      if (pwr > 0.99) pwr = 0.99;     // clamp. Do math in floating point to prevent any integer overflow
      if (pwr < 0.01) pwr = 0.01;     // it's slower. We could do clamp after conversion to int but it's your mosfets...
      vpwr = (int)((float)PPWM * pwr);// finally apply required PWM duty power factor to vpwr     
      check_switch_cont(); 
      }
}

void check_switch_cont()
  {   // called at 100 Hz
  char fbuf[20]; 
  int h,t;
  
  // inverter control
  if ((PINB & 0x01) == 0x01 && lv_stop == 0 && ot_stop == 0)     
    {
    if (init_done == 0)
      {
      init_vars();
      init_done = 1;
      }  
    oen = 1;      // slow start
    sbi(PORTB,4); // and light LED on D12, pin 15 of nano
    }
  if ((PINB & 0x01) == 0 || ot_stop == 1 || lv_stop == 1)       
    {
    init_done = 0;
    oen = 0;    // slow stop
    cbi(PORTB,4); // and blank LED on D12
    }   

  // fan control
  if (hs_temp > HS_STOP)        ot_stop = 1;
  if (tor_temp > TOR_STOP)        ot_stop = 1;
  if (NTC_COMBINE == 0)
    {
    if (hs_temp > HS_FAN_ON)      
      { 
        sbi(PORTD,3);       // heat sink fan 12V drive
      }
    if (hs_temp < HS_FAN_OFF)     
      {
        cbi(PORTD,3);
      }
    if (tor_temp > TOR_FAN_ON)      
      {
        sbi(PORTD,4);     // toroid fan drive
      }
    if (tor_temp < TOR_FAN_OFF)     
      {
        cbi(PORTD,4);
      }
    }
  else
    {
      if ((hs_temp > HS_FAN_ON) || (tor_temp > TOR_FAN_ON)) sbi(PORTD,3);
      if ((hs_temp < HS_FAN_OFF) && (tor_temp < TOR_FAN_OFF)) cbi(PORTD,3);
    }
    
  led_count++;
  if (ot_stop == 1 || lv_stop == 1)       // light Nano onboard LED if ot or lv
    {
      if (lv_stop == 1 && ot_stop == 0) digitalWrite(13, ((led_count & 0x40) == 0x40)? 1:0  ); 
      if (ot_stop == 1) digitalWrite(13, ((led_count & 0x20) == 0x20)? 1:0  );  
    }
    
  // check for LV condition, set lv stop flag = 1 if needed
  // needs to be true for entire period of LV_TIMEOUT
  if ( (bv < LV_CUT_OFF) && (LV_STOP_ENABLE == 1) )
    {
      lv_sec_count++;
      if (lv_sec_count >= 100)      // 100Hz means 100 counts per second
        {
          lv_time_count++;
          lv_sec_count = 0;
          if (lv_time_count > LV_TIMEOUT)
            lv_stop = 1;
        }
    }
  else
    {
      lv_sec_count = 0;
      lv_time_count = 0;
    }
  
  // restart after LV
  if (bv > LV_RESTART && LV_ENABLE)
    lv_stop = 0;  
            
  if (oen == 1)
    strncpy(bp,"ON  ",4);
  else
    strncpy(bp,"OFF ",4);
  if (ot_stop == 1)    strncpy(bp+20,"OT STOP ",8);  
  if (lv_stop == 1)    
    strncpy(bp+30,"LV STOP ",8);
  else
    strncpy(bp+30,"        ",8);
  h = t = ' ';
  if (PIND & 0x08)    h = '*';
  if (PIND & 0x10)    t = '*';  
  sprintf(bp+40,"HS %3d%c    TOR %3d%c  ",hs_temp,h,tor_temp,t);  
  dtostrf(bv,3,1,fbuf);
  sprintf(bp+60,"BV = %sV    ",fbuf);
  buf[73] = buf[74] = buf[75] = ' ';
  }


// should be fast
// clamp values below 0 C to 0 C and those over 100 C to 100 C
// this works only if NTC is on high voltage side of 10K divider
// and we use 10K NTC thermistors
// 
uint8_t lookup_temp(int i)
  {
    if (i < 270) i = 270;
    if (i > 943) i = 943;
    return ntc[i-270];
  }

//
// we will use a lookup table now...
// so need this for pre-filling the table
//
float get_temp(int Vo) 
 {    // https://circuitdigest.com/microcontroller-projects/arduino-thermistor-interfacing-code-circuit
 float a = 1.009249522e-03, b = 2.378405444e-04, c = 2.019202697e-07;
 float T,logRt,Tc;
 logRt = log(10000.0*((1024.0/Vo-1))); 
 T = (1.0 / (a + b*logRt + c*logRt*logRt*logRt)); // We get the temperature value in Kelvin from this Stein-Hart equation
 Tc = T - 273.15;                                 // Convert Kelvin to Celsius
 return Tc;
}
